/*
 *     AccessResolver.java  Mar 26, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.platform.j2xml;
import org.jtool.platform.xml.wrapper.*;
import org.jtool.platform.xml.JXConstants;
import org.w3c.dom.*;
import java.util.*;

public class AccessResolver extends SimpleResolver {
    private static Document document;
    private static HashMap fieldCache = new HashMap();  // String/JavaField
    private static HashMap methodCache = new HashMap();  // String/JavaMethod

    public static Document run(JavaXMLText text) {
        fieldCache.clear();
        methodCache.clear();

        document = text.getDocument();
        Element elem = document.getDocumentElement();
        elem.setAttribute(JXConstants.SortAttr, JXConstants.LEVEL3);

        NodeList nl = elem.getElementsByTagName(JXConstants.FieldElem);
        for (int i = 0; i < nl.getLength(); i++) {
            Element e = (Element)nl.item(i);
            String id = e.getAttribute(JXConstants.IdAttr);
            fieldCache.put(id, new JavaField(e));
        }

        nl = elem.getElementsByTagName(JXConstants.MethodElem);
        for (int i = 0; i < nl.getLength(); i++) {
            Element e = (Element)nl.item(i);
            String id = e.getAttribute(JXConstants.IdAttr);
            methodCache.put(id, new JavaMethod(e));
        }

        writeRefExpr(elem);

        return document;
    }

    private static void writeRefExpr(Node node) {
        NodeList nl = node.getChildNodes();
        for (int i = 0; i < nl.getLength(); i++) {
            Node child = nl.item(i);
            writeRefExpr(child);
        }

        if (node.getNodeType() == Node.ELEMENT_NODE) {
            if (node.getNodeName().equals(JXConstants.ExpressionElem)) {
                writeRefAttribute((Element)node);
            }
        }
    }
    
    private static void writeRefAttribute(Element e) {
        String sort = e.getAttribute(JXConstants.SortAttr);
        if (sort.equals(JXConstants.VarRef)) {
            JavaElement jelem = new JavaElement(e);
            Element ie = (Element)jelem.getFirstChildNode(e, JXConstants.IdentElem);
            String ref = ie.getAttribute(JXConstants.RefAttr);

            if (ref.length() == 0) {
                String defid = ie.getAttribute(JXConstants.DefIdAttr);

                if (defid.length() != 0) {
                    JavaField jf = (JavaField)fieldCache.get(defid);
                    if (jf != null) {
                        JavaVariable jv = new JavaVariable(e, jf.getElem());
                        writeRefAttribute(jv, jf);
                    }

                } else {
                    JavaVariable jv = new JavaVariable(e, null);
                    JavaField jf = getReference(jv);
                    writeRefAttribute(jv, jf);
                }
            }

        } else if (sort.equals(JXConstants.MethodCall)) {
            JavaElement jelem = new JavaElement(e);
            Element ie = (Element)jelem.getFirstChildNode(e, JXConstants.IdentElem);
            String ref = ie.getAttribute(JXConstants.RefAttr);
            
            if (ref.length() == 0) {
                String defid = ie.getAttribute(JXConstants.DefIdAttr);

                if (defid.length() != 0) {
                    JavaMethod jm = (JavaMethod)methodCache.get(defid);
                    if (jm != null) {
                        JavaCall ja = new JavaCall(e, jm.getElem());
                        writeRefAttribute(ja, jm);
                    }

                } else {
                    JavaCall ja = new JavaCall(e);
                    JavaMethod jm = getReference(ja);
                    writeRefAttribute(ja, jm);
                }
            }
        }
    }

    private static void writeRefAttribute(JavaElement je, String fqn, String ref, String defid) {
        Element elem = (Element)je.getFirstChildNode(je.getElem(), JXConstants.IdentElem);
        elem.setAttribute(JXConstants.FqnAttr, fqn);
        elem.setAttribute(JXConstants.RefAttr, ref);
        elem.setAttribute(JXConstants.DefIdAttr, defid);
        // System.out.println(je.getElem());
    }

    private static boolean isClassReference(Element prefix) {
        if (prefix != null &&
            prefix.getAttribute(JXConstants.SortAttr).equals(JXConstants.ClassRef)) {
            return true;
        }
        return false;
    }

    private static JavaField getReference(JavaVariable jv) {
        // System.out.println("FIELD REF = " + jv.getName() + " " + jv.getElemId());

        Element prefix = (Element)jv.getDotPrefix();
        String ptype = jv.getDotPrefixType(prefix);
        // System.out.println("PTYPE = " + ptype);

        String key = jv.getJavaClass().getFqn() + "#" + jv.getName() + "!";
        if (ptype != null) {
            key = key + ptype;
        }

        JavaField jf = (JavaField)fieldCache.get(key);
        if (jf == null) {
            if (ptype != null) {
                if (ptype.endsWith("[]")) {
                    jf = new JavaArrayField(jv.getName());
                
                } else {
                    if (isClassReference(prefix)) {
                        jf = getAccessFieldInClass(ptype, jv);

                    } else {
                        jf = getAccessFieldInAncestors(ptype, jv);
                    }                    
                }
                
            } else {
                JavaClass jc = jv.getJavaClass();
                if (jc.isAnonymous()) {
                    jf = getAccessFieldInAncestors(jc.getFqnOfAnonymous(), jv);
                } else {
                    jf = getAccessFieldInAncestors(jc.getFqn(), jv);
                }
            
                while (jf == null && jc.isInnerClass()) {
                    jc = new JavaClass(jc.getOuterClassNode());
                        
                    if (jc.isAnonymous()) {
                        jf = getAccessFieldInAncestors(jc.getFqnOfAnonymous(), jv);
                    } else {
                        jf = getAccessFieldInAncestors(jc.getFqn(), jv);
                    }
                }
            }
        }

        if (jf != null && fieldCache.get(key) == null) {
            fieldCache.put(key, jf);
        }
        return jf;
    }

    private static void writeRefAttribute(JavaVariable jv, JavaField jf) {
        if (jf != null) {
            String fqn = jv.getCastType();
            if (fqn == null) {
                fqn = jf.getTypeFqn();
            }
            String ref = jf.getJavaClass().getFqn();
            String defid = jf.getId();

            writeRefAttribute(jv, fqn, ref, defid);
            setAncestorsOfClass(jv.getTypeFqn());
            // System.out.println("FIELD ACCESS = " + jv.getName() + "[" + jv.getTypeFqn() + "]: " +
            //  jv.getRefClass() + "/" + defid);
            // System.out.println(" " + jv.getUsedJavaField());

        } else {
            // System.out.println("Invalid field access = " + jv.getName() + " " + jv.getJavaFile().getName());
        }
    }

    private static JavaField getAccessFieldInAncestors(String type, JavaVariable jv) {
        // System.out.println("JV_TYPE = " + type + " " + jv.getName());

        if (type != null && type.length() != 0) {
            JavaClass jclass = Repository.getJavaClass(type, jv.getJavaFile());

            if (jclass != null) {
                JavaClassList classes = getTraverseClasses(jclass);
                return getAccessField(classes, jv);
            }
        }
        return null;
    }

    private static JavaField getAccessFieldInClass(String type, JavaVariable jv) {
        if (type != null) {
            JavaClass jclass = Repository.getJavaClass(type, jv.getJavaFile());

            JavaClassList classes = new JavaClassList();
            classes.add(jclass);
            return getAccessField(classes, jv);
        }
        return null;
    }

    private static JavaField getAccessField(JavaClassList classes, JavaVariable jv) {
        Iterator it = classes.iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();
            if (jc != null) {
                JavaField jf = jc.getJavaField(jv.getName());
                if (jf != null && isAccessible(jc, jf)) {
                    return jf;
                }
            }
        }
        return null;
    }

    private static boolean isAccessible(JavaClass jc, JavaField jf) {
        JavaClass c = jf.getJavaClass();

        if (jc.equals(c) || jf.isPublic()) {
            return true;
        }

        JavaClassList ancestors = AncestorResolver.getAncestors(jc);
        if (jf.isProtected() && ancestors.contains(c)) {
            return true;
        }

        if (!jf.isPrivate() && jc.isInTheSamePackage(c)) {
           return true;
        }

        return false;
    }

    private static JavaMethod getReference(JavaCall ja) {
        // System.out.println("METHOD CALL = " + ja.getName() + " " + ja.getElemId());

        Element prefix = (Element)ja.getDotPrefix();
        String ptype = ja.getDotPrefixType(prefix);
        // System.out.println("PTYPE = " + ptype);
        
        String key = ja.getJavaClass().getFqn() + "#" + ja.getSignature() + "!";
        if (ptype != null) {
            key = key + ptype;
        }

        JavaMethod jm = (JavaMethod)methodCache.get(key);
        if (jm == null) {
            if (ptype != null) {
                if (ptype.endsWith("[]")) {
                    jm = new JavaArrayMethod(ja.getSignature());

                } else {
                    if (isClassReference(prefix)) {
                        jm = getAccessMethodInClass(ptype, ja);

                    } else {
                        jm = getAccessMethodInAncestors(ptype, ja);
                    }
                }

            } else {
                JavaClass jc = ja.getJavaClass();
                if (jc.isAnonymous()) {
                    jm = getAccessMethodInAncestors(jc.getFqnOfAnonymous(), ja);
                } else {
                    jm = getAccessMethodInAncestors(jc.getFqn(), ja);
                }
            
                while (jm == null && jc.isInnerClass()) {
                    jc = new JavaClass(jc.getOuterClassNode());

                    if (jc.isAnonymous()) {
                        jm = getAccessMethodInAncestors(jc.getFqnOfAnonymous(), ja);
                    } else {
                        jm = getAccessMethodInAncestors(jc.getFqn(), ja);
                    }
                }
            }
        }

        if (jm != null && methodCache.get(key) == null) {
            methodCache.put(key, jm);
        }
        return jm;
    }

    private static void writeRefAttribute(JavaCall ja, JavaMethod jm) {
        if (jm != null) {
            String fqn = ja.getCastType();
            if (fqn == null) {
                fqn = jm.getTypeFqn();
            }
            String ref = jm.getJavaClass().getFqn();
            String defid = jm.getId();

            writeRefAttribute(ja, fqn, ref, defid);
            setAncestorsOfClass(ja.getTypeFqn());
            // System.out.println("METHOD ACCESS = " + ja.getName() + "[" + ja.getTypeFqn() + "]: " +
            //                    ja.getRefClass() + "/" + defid);
            // System.out.println(" " + ja.getCalledJavaMethod());

        } else {
            // System.out.println("Invalid method call = " + ja.getName() + " " + ja.getJavaFile().getName());
        }
    }

    private static JavaMethod getAccessMethodInAncestors(String type, JavaCall ja) {
        // System.out.println("JA_TYPE = " + type + " " + ja.getName() + " " + ja.getNumArgument());

        if (type != null && type.length() != 0) {
            JavaClass jclass = Repository.getJavaClass(type);

            if (jclass != null) {
                JavaClassList classes = getTraverseClasses(jclass);
                return getAccessMethod(classes, ja);
            }
        }
        return null;
    }

    private static JavaMethod getAccessMethodInClass(String type, JavaCall ja) {
        if (type != null) {
            JavaClass jclass = Repository.getJavaClass(type);

            JavaClassList classes = new JavaClassList();
            classes.add(jclass);
            return getAccessMethod(classes, ja);
        }
        return null;
    }

    private static JavaMethod getAccessMethod(JavaClassList classes, JavaCall ja) {
        JavaMethodList candidates = new JavaMethodList();

        Iterator it = classes.iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();
            if (jc != null) {
                candidates.addAll(collectCandidateMethods(jc, ja));
            }
        }

        if (candidates.size() > 1) {
            candidates = removeMethodsWithSameName(candidates);
            candidates = removeLessSpecificMethods(candidates);
        }

        if (candidates.size() == 1) {
            JavaMethod jm = (JavaMethod)candidates.get(0);
            return jm;

        } else {
            // System.out.println("Invalid method dispatch");
            // printMethods(candidates);
        }
        return null;
    }

    private static void printClasses(JavaClassList classes) {
        Iterator it = classes.iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();
            System.out.println("  CHECK CLASS = " + jc.getFqn());
        }
    }

    private static void printMethods(JavaMethodList methods) {
        Iterator it = methods.iterator();
        while (it.hasNext()) {
            JavaMethod jm = (JavaMethod)it.next();
            System.out.println("   CANDIDATE = " + jm.getSignature());
        }
    }

    private static JavaMethodList collectCandidateMethods(JavaClass jc, JavaCall ja) {
        JavaMethodList candidates = new JavaMethodList();

        if (jc != null) {
            Iterator it = jc.getMethodNodes().iterator();
            while (it.hasNext()) {
                JavaMethod jm = new JavaMethod((Node)it.next());
                if (ja.getName().equals(jm.getName())) {
                    if (isAccessible(jc, jm) && isApplicable(ja, jm)) {
                        candidates.add(jm);
                    }
                }                    
            }

            it = jc.getConstructorNodes().iterator();
            while (it.hasNext()) {
                JavaMethod jm = new JavaMethod((Node)it.next());
                if (isAccessible(jc, jm) && isApplicable(ja, jm)) {
                    candidates.add(jm);
                }
            }
        }

        // printMethods(candidates);
        return candidates;
    }

    private static boolean isAccessible(JavaClass jc, JavaMethod jm) {
        JavaClass c = jm.getJavaClass();

        if (jc.equals(c) || jm.isPublic()) {
            return true;
        }

        if (!jm.isPrivate() && jc.isInTheSamePackage(c)) {
            return true;
        }
        
        JavaClassList ancestors = AncestorResolver.getAncestors(jc);
        if (jm.isProtected() && ancestors.contains(c)) {
            return true;
        }

        return false;
    }

    private static boolean isApplicable(JavaCall ja, JavaMethod jm) {
        // System.out.println("JA = " + ja.getSignature());
        // System.out.println("JM = " + jm.getSignature());

        if (ja.getSignature().equals(jm.getSignature())) {
            return true;
        }
        if (ja.getNumArgument() == jm.getNumParameter() &&
            ja.getName().equals(jm.getName()) && isConvertible(ja, jm)) {
            return true;
        }
        return false;
    }        

    private static boolean isConvertible(JavaCall ja, JavaMethod jm) {
        for (int i = 0; i < ja.getNumArgument(); i++) {
            String argu = ja.getArgumentTypeFqn(i);
            String para = jm.getParameterTypeFqn(i);

            if (!isConvertible(argu, para)) {
                return false;
            }
        }
        return true;
    }

    private static boolean isConvertible(String tp, String up) {
        if (tp == null || up == null) {
            return false;
        }

        if (tp.equals(up) || isSpeciallyConvertible(tp, up)) {
            return true;
        }

        while (tp.endsWith("[]") && up.endsWith("[]")) {
            tp = tp.substring(0, tp.lastIndexOf("["));
            up = up.substring(0, up.lastIndexOf("["));
        }
        if (tp.endsWith("[]") || up.endsWith("[]")) {
            return false;
        }

        if (TypeConversion.isPrimitiveType(tp) && TypeConversion.isPrimitiveType(up)) {
            return isConvertiblePrimitiveType(tp, up);
        }

        if (!TypeConversion.isPrimitiveType(tp) && !TypeConversion.isPrimitiveType(up)) {
            JavaClass tc = Repository.entryClass(tp);
            JavaClass uc = Repository.entryClass(up);
            return isConvertibleClassType(tc, uc);
        }
        return false;
    }

    private static boolean isSpeciallyConvertible(String tp, String up) {
        if (tp.equals("null")) {
            return true;
        }

        if (tp.endsWith("[]")) {
            if (up.equals("java.lang.Object") ||
                up.equals("java.lang.Cloneable") ||
                up.equals("java.lang.Serializable")) {
                return true;
            }
        }

        if (!TypeConversion.isPrimitiveType(tp) && up.equals("java.lang.Object")) {
            return true;
        }

        return false;
    }

    private static boolean isConvertiblePrimitiveType(String tp, String up) {
        String type = TypeConversion.wideningPrimitiveType(tp, up);
        if (type.equals(up)) {
            return true;
        }
        return false;
    }

    private static boolean isConvertibleClassType(JavaClass tc, JavaClass uc) {
        if (tc != null && uc != null) {
            if (tc.equals(uc)) {
                return true;
            }

            if (uc.getFqn().equals("java.lang.Object")) {
                return true;
            }

            JavaClassList ancestors = AncestorResolver.getAncestors(tc);
            if (ancestors.contains(uc)) {
                return true;
            }
        }
        return false;
    }

    private static JavaMethodList removeLessSpecificMethods(JavaMethodList candidates) {
        JavaMethodList tmpCandidates = new JavaMethodList(candidates);

        for (int i = 0; i < tmpCandidates.size(); i++) {
            JavaMethod jm1 = (JavaMethod)tmpCandidates.get(i);

            for (int j = 0; j < tmpCandidates.size(); j++) {
                JavaMethod jm2 = (JavaMethod)tmpCandidates.get(j);

                if (i != j && isMoreSpecific(jm1, jm2)) {
                    candidates.remove(jm2);
                }
            }
        }
        return candidates;
    }

    private static boolean isMoreSpecific(JavaMethod jm1, JavaMethod jm2) {
        JavaClass jc1 = jm1.getJavaClass();
        JavaClass jc2 = jm2.getJavaClass();

        JavaClassList ancestors = AncestorResolver.getAncestors(jc1);
        if (!jc1.equals(jc2) && !ancestors.contains(jc2)) {
            return false;
        }
        return isConvertible(jm1, jm2);
    }

    private static boolean isConvertible(JavaMethod jm1, JavaMethod jm2) {
        for (int i = 0; i < jm1.getNumParameter(); i++) {
            String para1 = jm1.getParameterTypeFqn(i);
            String para2 = jm2.getParameterTypeFqn(i);

            if (!isConvertible(para1, para2)) {
                return false;
            }
        }
        return true;
    }

    private static JavaMethodList removeMethodsWithSameName(JavaMethodList candidates) {
        JavaMethodList tmpCandidates = new JavaMethodList(candidates);

        for (int i = 0; i < tmpCandidates.size(); i++) {
            JavaMethod jm1 = (JavaMethod)tmpCandidates.get(i);

            for (int j = i + 1; j < tmpCandidates.size(); j++) {
                JavaMethod jm2 = (JavaMethod)tmpCandidates.get(j);

                if (jm1.getSignature().equals(jm2.getSignature())) {
                    candidates.remove(jm2);
                }
            }
        }
        return candidates;
    }

    private static JavaClassList getTraverseClasses(JavaClass jclass) {
        JavaClassList bases = new JavaClassList();
        bases.add(jclass);

        if (jclass.isInnerClass()) {
            Iterator it = jclass.getOuterClassNodes().iterator();
            while (it.hasNext()) {
                JavaClass jc = new JavaClass((Node)it.next());
                bases.add(jc);
            }
        }
        
        JavaClassList classes = new JavaClassList();
        Iterator it = bases.iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();
            classes.addAll(AncestorResolver.getAncestors(jclass));
        }
        return classes;
    }

    private static void setAncestorsOfClass(String fqn) {
        if (fqn != null) {

            String fqnPrefix = fqn;
            int sep = fqnPrefix.indexOf("[");
            if (sep != -1) {
                fqnPrefix = fqnPrefix.substring(0, sep);
            }

            if (!TypeConversion.isPrimitiveType(fqnPrefix)) {
                JavaClass jc = Repository.entryClass(fqnPrefix);
                
                if (jc != null) {
                    String pathName = jc.getJavaFile().getName();
                    ClassXMLText text = Repository.getClassXMLText(pathName);

                    if (text != null) {
                        AncestorResolver.setAncestors(text);
                        TypeResolver.writeResolvedFqn(text);
                    }
                }
            }
        }
    }
}
