/*
 *     GenFileXML.java  Jan 30, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.platform.xml.xref;
import org.jtool.platform.j2xml.*;
import org.jtool.platform.util.*;
import org.jtool.platform.xml.wrapper.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;

public class GenFileXML {
    private static StringBuffer buf;
    private static String thisFileName; 

    public static void write() {
        write(Repository.getJavaTexts());
    }

    public static void write(List texts) {
        Iterator it = texts.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();

            JavaFile jfile = new JavaFile(text.getDocument().getDocumentElement());
            thisFileName = getPath(jfile.getName());

            buf = new StringBuffer();
            writeFile(jfile);

            try {
                String xmlName = GetXref.getPgStructFileName(jfile.getName());

                String dirName = xmlName.substring(0, xmlName.lastIndexOf(File.separator));
                SystemProperties.makeDir(dirName);

                WriteFile.write(xmlName, buf.toString());
                System.out.println("WRITE: " + xmlName);
            } catch (IOException e) { }
        }
    }

    private static void writeFile(JavaFile jfile) {
        buf.append("<" + Constants.FileElem);
        buf.append(" " + Constants.IdAttr + "=\"" + jfile.getId() + "\"");
        buf.append(">\n");

        buf.append("  <" + Constants.AnchorElem);
        buf.append(" " + Constants.HrefAttr + "=\"" + thisFileName + "\"");
        buf.append(" " + Constants.LineAttr + "=\"" + jfile.getLine() + "\"");
        buf.append(">");
        buf.append(jfile.getBaseName());
        buf.append("</" + Constants.AnchorElem + ">\n");

        writePackage(jfile);
        writeImport(jfile);

        writeClass(jfile);

        writeClassReference(jfile);
        writeFieldReference(jfile);
        writeMethodReference(jfile);

        buf.append("</" + Constants.FileElem + ">\n");
    }
        
    private static void writePackage(JavaFile jfile) {
        if (jfile.getPackageName().length() != 0) {
            buf.append("<" + Constants.PackageElem);
            JavaPackage jp = new JavaPackage(jfile.getPackageNode());

            buf.append(" " + Constants.IdAttr + "=\"" + jp.getId() + "\"");
            buf.append(">");

            buf.append("<" + Constants.AnchorElem);
            buf.append(" " + Constants.HrefAttr + "=\"" + thisFileName + "\"");
            buf.append(" " + Constants.LineAttr + "=\"" + jp.getLine() + "\"");
            buf.append(">");
            buf.append(jp.getName());
            buf.append("</" + Constants.AnchorElem + ">");

            buf.append("</" + Constants.PackageElem + ">\n");
        }
    }

    private static void writeImport(JavaFile jfile) {
        Iterator it = jfile.getImportNodes().iterator();
        while (it.hasNext()) {
            JavaImport ji = new JavaImport((Element)it.next());

            buf.append("  <" + Constants.ImportElem);
            buf.append(" " + Constants.IdAttr + "=\"" + ji.getId() + "\"");
            buf.append(">");

            buf.append("<" + Constants.AnchorElem);
            buf.append(" " + Constants.HrefAttr + "=\"" + thisFileName + "\"");
            buf.append(" " + Constants.LineAttr + "=\"" + ji.getLine() + "\"");
            buf.append(">");
            buf.append(ji.getName());
            buf.append("</" + Constants.AnchorElem + ">");

            buf.append("</" + Constants.ImportElem + ">\n");
        }
    }

    private static void writeClass(JavaFile jfile) {
        Iterator it = jfile.getAllClasses().iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();
            GenClassXML.writeClass(jc, buf, "  ");
        }

        it = jfile.getAllInterfaces().iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();
            GenClassXML.writeClass(jc, buf, "  ");
        }
    }

    private static void writeClassReference(JavaFile jfile) {
        Iterator it = jfile.getTypeNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            JavaClass jc = ReferenceUtil.getRefJavaClass(elem);
            writeReference(new JavaType(elem), new RefJavaClass(jc), Constants.ClassRefElem);
        }
    }

    private static void writeFieldReference(JavaFile jfile) {
        Iterator it = jfile.getExpressionNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            JavaField jf = ReferenceUtil.getRefJavaField(elem);
            writeReference(new JavaVariable(elem), new RefJavaField(jf), Constants.FieldRefElem);
        }
    }

    private static void writeMethodReference(JavaFile jfile) {
        Iterator it = jfile.getExpressionNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            JavaMethod jm = ReferenceUtil.getRefJavaMethod(elem);
            writeReference(new JavaCall(elem), new RefJavaMethod(jm), Constants.MethodRefElem);
        }
    }

    private static void writeReference(JavaElement je, RefJavaElement rje, String elemName) {
        if (rje != null)  {
            buf.append("  <" + elemName);
            buf.append(" " + Constants.IdAttr + "=\"" + je.getId() + "\"");
            buf.append(">\n");

            buf.append("    <" + Constants.AnchorElem);
            buf.append(" " + Constants.HrefAttr + "=\"" + thisFileName + "\"");
            buf.append(" " + Constants.LineAttr + "=\"" + je.getLine() + "\"");
            buf.append(">");
            buf.append(rje.getName());  // truly je.getName()
            buf.append("</" + Constants.AnchorElem + ">\n");

            buf.append("    <" +  Constants.RefTargetElem);
            buf.append(" " + Constants.IdAttr + "=\"" + rje.getId() + "\"");
            buf.append(" " + Constants.FqnAttr + "=\"" + rje.getFqn() + "\"");
            if (GetXref.isInProject(rje.getJavaClass())) {
                buf.append(" " + Constants.ExistsAttr + "=\"" + Constants.InProject + "\"");
            } else if (GetXref.isInJDK(rje.getJavaClass())) {
                buf.append(" " + Constants.ExistsAttr + "=\"" + Constants.InJDK + "\"");
            }
            buf.append(">\n");

            buf.append("      <" + Constants.AnchorElem);
            buf.append(" " + Constants.HrefAttr + "=\"" + getPath(rje.getJavaFile().getName()) + "\"");
            buf.append(" " + Constants.LineAttr + "=\"" + rje.getLine() + "\"");

            buf.append(">");
            buf.append(rje.getName());
            buf.append("</" + Constants.AnchorElem + ">\n");
            buf.append("    </" +  Constants.RefTargetElem + ">\n");

            buf.append("  </" + elemName + ">\n");
        }  
    }

    private static String getPath(String path) {
        return SystemProperties.getRelativePath(path, SystemProperties.getSourceDir());
    }
}
