/*
 *     ReferenceUtil.java  Jan 30, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.platform.xml.xref;
import org.jtool.platform.xml.wrapper.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;

public class ReferenceUtil {
    private static ArrayList id2je = new ArrayList();

    public static void makeId2JavaElement() {
        id2je.clear();

        Iterator it = Project.getAllFiles().iterator();
        while (it.hasNext()) {
            JavaFile jfile = (JavaFile)it.next();

            collectClassReference(jfile);
            collectFieldReference(jfile);
            collectMethodReference(jfile);
        }
    }     

    private static void collectClassReference(JavaFile jfile) {
        Iterator it = jfile.getTypeNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            JavaClass rjc = getRefJavaClass(elem);
            entryReference(new JavaType(elem), rjc);
        }
    }

    private static void collectFieldReference(JavaFile jfile) {
        Iterator it = jfile.getExpressionNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            JavaField rjf = getRefJavaField(elem);
            entryReference(new JavaVariable(elem), rjf);
        }
    }
    
    private static void collectMethodReference(JavaFile jfile) {
        Iterator it = jfile.getExpressionNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            JavaMethod rjm = getRefJavaMethod(elem);
            entryReference(new JavaCall(elem), rjm);
        }
    }

    public static JavaClass getRefJavaClass(Element elem) {
        JavaType jt = new JavaType(elem);
        if (jt.isObjectRef() || jt.isClassRef()) {
            return jt.getReferredJavaClass();
        }
        return null;
    }

    public static JavaField getRefJavaField(Element elem) {
        JavaExpr jr = new JavaExpr(elem);
        if (jr.isVarRef()) {
            JavaVariable jv = new JavaVariable(jr.getElem());
            return jv.getUsedJavaField();
        }
        return null;
    }
    
    public static JavaMethod getRefJavaMethod(Element elem) {
        JavaExpr jr = new JavaExpr(elem);
        if (jr.isMethodCall() || jr.isConstructorCall() || jr.isSuperConstructorCall()) {
            JavaCall ja = new JavaCall(jr.getElem());
             return ja.getCalledJavaMethod();
        }
        return null;
    }

    private static void entryReference(JavaElement je, JavaElement rje) {
        if (rje != null) {
            String fileName = rje.getJavaFile().getName();
            String id = fileName + File.pathSeparator + rje.getId();
            if (GetXref.isInProject(fileName)) {
                id2je.add(new Id2JavaElement(id, je));
            }
        }
    }

    public static ArrayList getReferringNodes(String id) {
        ArrayList nodes = new ArrayList();

        Iterator it = id2je.iterator();
        while (it.hasNext()) {
            Id2JavaElement e = (Id2JavaElement)it.next();
            
            if (e.getId().equals(id)) {
                nodes.add(e.getJavaElement());
            }
        }
        return nodes;
    }
}

class Id2JavaElement {
    private String id;
    private JavaElement je;

    Id2JavaElement(String id, JavaElement je) {
        this.id = id;
        this.je = je;
    }

    String getId() {
        return id;
    }

    JavaElement getJavaElement() {
        return je;
    }
}
